#!/usr/bin/env python
#
# alz2zip.py
#
# Allows to convert simple Alzip files into Zip file.
#
# ----------------------------------------------------------------------------
# "THE BEER-WARE LICENSE" (Revision 42, (c) Poul-Henning Kamp):
# Hye-Shik Chang <perky@FreeBSD.org> wrote this file.  As long as you retain
# this notice you can do whatever you want with this stuff. If we meet some
# day, and you think this stuff is worth it, you can buy me a beer in return.
#
# Hye-Shik Chang
# ----------------------------------------------------------------------------
#
# $Perky: snippets/alz2zip/alz2zip.py,v 1.1.1.1 2004/02/16 15:14:20 perky Exp $

from struct import *

__all__ = ['convert', 'AlzipError']
__version__ = '0.1'

SIG_ALZ = 0x015A4C41
SIG_BLZ = 0x015A4C42
SIG_CLZ = 0x015A4C43
SIG_ZIP_FILE_HEADER = 0x04034B50
SIG_ZIP_CENTRAL_DIRECTORY = 0x02014B50
SIG_ZIP_END_OF_CENTRAL_DIRECTORY_RECORD = 0x06054B50

class AlzipError(Exception): pass

class Alz2ZipConvertSession(object):

    def __init__(self, infile, outfile):
	self.infile = infile
	self.outfile = outfile
	self.convert()

    def safeunpack(self, fmt, length):
	expsize = calcsize(fmt)
	return unpack(fmt, self.read(length) + '\x00' * (expsize - length))

    def convert(self):
	headerinfo = self.read_file_header()
	if headerinfo['signature'] != SIG_ALZ:
	    raise AlzipError, 'infile is not an alzip file.'

	datainfo = self.read_file_data()
	if datainfo['data_signature'] != SIG_BLZ:
	    raise AlzipError, 'no compressed file found!'

	zipinfo = {
	'ver_needed': 0x0014,
	'gpb_flag': 0x0004,
	'method': datainfo['compression_rate'] and 8,
	'time': datainfo['time'],
	'date': datainfo['date'],
	'crc32': datainfo['crc32'],
	'filename_length': datainfo['filename_length'],
	'extra_field_length': 0,
	'compressed_size': self.safeunpack('<I',
				datainfo['compression_info'] >> 4)[0],
	'uncompressed_size': self.safeunpack('<I',
				datainfo['compression_info'] >> 4)[0],
	}
	filename = self.read(datainfo['filename_length'])
	self.write(pack('<I', SIG_ZIP_FILE_HEADER))
	self.write(self.pack_zip_file_info(zipinfo))
	self.write(filename)

	leftbytes = zipinfo['compressed_size']
	while leftbytes > 0:
	    data = self.read(min(8192, leftbytes))
	    leftbytes -= len(data)
	    self.write(data)

	zip_centraldir_info = {
	'signature': SIG_ZIP_CENTRAL_DIRECTORY,
	'ver_made': 0x0014,
	'compressed_file': self.pack_zip_file_info(zipinfo),
	'file_comment_length': 0,
	'disk_number_start': 0,
	'internal_file_attributes': datainfo['attributes'] & 0xff,
	'external_file_attributes': datainfo['attributes'],
	'relative_offset': 0,
	}
	zip_centraldir_packed = self.pack_zip_centraldir_info(
				    zip_centraldir_info)
	zip_centraldir_offset = self.outfile.tell()
	self.write(zip_centraldir_packed)
	self.write(filename)

	zip_dir_record_info = {
	'signature': SIG_ZIP_END_OF_CENTRAL_DIRECTORY_RECORD,
	'num_of_disk': 0,
	'central_directory': 1,
	'total_num_of_entries': 1,
	'size_of_central_directory': len(zip_centraldir_packed) +
				     zipinfo['filename_length'],
	'offset_of_central_directory': zip_centraldir_offset,
	'zipfile_comment_length': 0,
	}
	self.write(self.pack_zip_dir_record(zip_dir_record_info))

    def read_by_fmt(self, fmt, pushpadbytes=[]):
	expsize = calcsize(fmt) - len(pushpadbytes)
	data = self.read(expsize)
	if len(data) < expsize:
	    raise AlzipError, 'EOF found on reading a data chunk.'
	if pushpadbytes:
	    data = list(data)
	    for offset in pushpadbytes:
		data.insert(offset, '\x00')
	    data = str.join('', data)
	return unpack(fmt, data)

    def read_file_header(self):
	return dict(zip(('signature', 'version'),
			 self.read_by_fmt('<II')))
    
    def read_file_data(self):
	return dict(zip(('data_signature', 'filename_length',
			 'attributes', 'time', 'date',
			 'compression_info', 'compression_rate',
			 'crc32'), self.read_by_fmt('<IHHHHHHI', [7])))

    def pack_zip_file_info(self, info):
	fields = ('ver_needed', 'gpb_flag', 'method', 'time', 'date', 'crc32',
		  'compressed_size', 'uncompressed_size', 'filename_length',
		  'extra_field_length')
	return pack('<HHHHHIIIHH', *[info[f] for f in fields])

    def pack_zip_centraldir_info(self, info):
	fields1 = ('signature', 'ver_made')
	fields2 = ('file_comment_length', 'disk_number_start',
		   'internal_file_attributes', 'external_file_attributes',
		   'relative_offset')
	return (pack('<IH', *[info[f] for f in fields1]) +
		info['compressed_file'] +
		pack('<HHHII', *[info[f] for f in fields2]))

    def pack_zip_dir_record(self, info):
	fields = ('signature', 'num_of_disk', 'central_directory',
		  'total_num_of_entries', 'size_of_central_directory',
		  'offset_of_central_directory', 'zipfile_comment_length')
	return pack('<IIHHIIH', *[info[f] for f in fields])

    def read(self, *arg, **kw): return self.infile.read(*arg, **kw)
    def write(self, *arg, **kw): return self.outfile.write(*arg, **kw)

def convert(infile, outfile):
    Alz2ZipConvertSession(infile, outfile)

def main(argv):
    import sys, getopt
    try:
	opts, args = getopt.getopt(argv[1:], 'hx')
    except getopt.error, why:
	print >> sys.stderr, '%s: %s' % (argv[0], why)

    if not args:
	opts = (('-h', ''),)

    for opt, arg in opts:
	if opt == '-h':
	    print """\
alz2zip %s, a zip archive file sanitizer
Usage: %s [OPTION]... [URL]...
 
  -h     print this help.
 
Mail bug reports and patches to <perky@FreeBSD.org>.\
""" % (__version__, argv[0])
	    raise SystemExit

    for afile in args:
	try:
	    infile = open(afile)
	except IOError, why:
	    print >> sys.stderr, "%s: %s" % (argv[0], why)

	print '>>> Converting %s ...' % afile

	if afile.lower().endswith('alz'):
	    outfile = open(afile[:-3] + 'zip', 'w')
	else:
	    outfile = open(afile + '.zip', 'w')

	convert(infile, outfile)

if __name__ == '__main__':
    import sys
    main(args)

# ex: ts=8 sts=4 sw=4 noet
